Title:   Ubuntu Core Product Integration API Guide
Version: 20151105

[TOC]


Definitions
============

-   **Manifest**: used to define a Product
    that is based on Snaps. This is optional for image-based workflows.
-   **Image Reference**: information passed
    to the Provisioning Kit to use in setting up a device for testing
    the Product. It is free-form with data of your choosing (e.g. a
    pointer to a binary to be provisioned on the device). Defined either
    via a manifest or advertised to the system on-demand.
-   **Test Spec**: defines Tests to be run on
    matching images
-   **Test Opportunity**: an instance of a
    test for a given image and/or set of snaps that will be delivered to an Agent.
-   **Agent**: Client software installed on a host computer to run
    tests against a target device.
-   **Test Result**: metadata about output of a test run.

Authentication
============

All API endpoints are authenticated, which ensures private access
to your Organization's information.

To authenticate properly to the system, you'll first need oauth
credentials that are obtained via the Agent (see next chapter below,
“Lab Operation”, under “<span class="c28">[Authentication](#h.u7ke0hug9g6j)”
section).

<!-- LINK TO CORRECT PLACE IN TUTORIAL? -->

With those credentials, you can use an oauth library in your
language of choice to sign HTTP requests to the API endpoints.


We'll walk through an example of that in Python using <span
class="c6">requests\_oauthlib library that is installed
along with the Agent. "consumer\_key", "consumer\_secret", "token\_key",
and "token\_secret" should be replaced with your credentials.


In the Agent directory, activate the virtualenv:

    $ . env/bin/activate

Start a python interpreter:

    $ python

Import the the requests/oauth lib:

    >>> import requests_oauthlib

Setup vars with our credentials:

    >>> consumer_key = 'tophopstop'
    >>> consumer_secret = 'supersecretlongstringfullofsecretdata'
    >>> token_key = 'hethreetreebee'
    >>> token_secret = 'supersecretlongstringfullofsecretdata'

Obtain an authenticated session, this will sign all subsequent
requests made with that session using our credentials:

    >>> authenticated_session = requests_oauthlib.OAuth1Session(
    ... consumer_key,
    ... client_secret=consumer_secret,
    ... resource_owner_key=tokenkey,
    ... resource_owner_secret=token_secret)

Make a signed request to the Manifest Listing endpoint:

    >>> url = "https://spi.canonical.com/orgs/docs-org/products/manifests"
    >>> res = authenticated_session.get(url)



The authenticated\_session will oauth-sign the request, appending
an HTTP Authorization header that looks like:

    Authorization: OAuth oauth_nonce="...",
    oauth_timestamp="1440687597", oauth_version="1.0",
    oauth_signature_method="HMAC-SHA1", oauth_consumer_key="...",
    oauth_token="...", oauth_signature="..."

We can now check the response code and data:

    >>> res.status_code
    200
    >>> res.json()
    {'manifests': [{'base\_image\_reference': 'http://foo.com/bar',
        'created_at': '2015-08-25T08:00:10.500533',
        'id': '71e81dbb-4eb5-45df-a813-6e9efe0851be',
        'image_name': 'docs-img68f9005f7fdb4c07ae322de432f22f26',
        'release': '15.04',
        'snaps': [{'locked': True, 'name': 'snap-a', 'revision': 1},
                  {'locked': False, 'name': 'snap-b', 'revision': None}]}]}

In order to simplify the examples given in the API references
throughout this document, we have provided a helper script as part of
the Agent release called "api\_example.py" that has a basic curl-like
command-line interface. It will read the authentication credentials from
a provided config file and sign requests on your behalf. The examples
assume that you have placed your credentials (as output from the Agent
login process) into a file called conf.ini. Note, this script is meant only as an example and not
intended for production use.

Errors
============

The APIs use standard HTTP response codes. Most errors will
contain a human-readable "message" field in the JSON response body. Here
are the most common error codes and guidelines for handling them.

<table  class="table">
<tr>
    <th>Code        <th>Notes
<tr>
    <td><code>400</code>    <td>The request data (either POST body or GET query parameters) were invalid. Fix the data and retry the request.
<tr>
    <td><code>401</code>    <td>"message":  "invalid oauth signature"

        <p>Check that your  OAuth credentials are configured correctly, otherwise try obtaining new credentials with the Agent interactive login.
        <p>See also: https://forms.canonical.com/sso-support
<tr>
    <td><code>403</code>    <td>Access is forbidden for the requested resource. This generally means that the requested resource does not belong to the organization of the requesting user. Check that you are using the correct org-id and authentication credentials.
<tr>
    <td><code>404</code>    <td>A request was  made for a resource that does not exist. For example requesting manifest details for a mistyped manifest-id. Check that the the entity exists via the appropriate listing endpoint and try again.
<tr>
    <td><code>50x</code>    <td>Any of 500, 502, 503, 504.

        <p>There was an temporary error on the server. The request may be retried after a short wait. If the problem persists please contact support. Scripts should use an exponential backoff to avoid hammering the servers with retries.
</table>

API Reference
=============

Product Management
------------------

Note: Product Manifests support products built from
Snaps. If you are instead bootstrapping Product Integration with a pre-existing binary
image, please skip ahead to the next section - Test Management.

A Product is defined by a Manifest that has an image\_name
and either a list of snaps that comprise the image (e.g. as specified in
a Gadget Snap) or a base\_image\_reference.

The image\_name is used to relate Test Specs (see below) to
Manifests.

Manifests and Test Specs may be deactivated or deleted
permanently. Even if deleted, details of the Test Spec and Manifest (if
any) that led to the Test Opportunity are persisted in the Test Result
data.



To modify a Test Spec or Manifest, the user should delete the
existing one and create a new one with the desired changes. That will
not automatically trigger new tests, for that see [Test Triggering](#test-triggering).

A manifest contains the following fields:

-   ```image_name```: a unique name within an
    organization that is combined with each
    test case to produce the full set of test
    opportunities. I.e. if you have two test cases with a
    matching image\_name, then two test opportunities will be created.
    Maximum length is 200 characters.
-   ```release``` snappy release series, e.g.
    “15.04”. Maximum length is 200 characters.
-   ```base_image_reference``` is opaque to
    the system; it is for your Provisioning Kit in your lab to
    interpret and use. It can be any JSON
    value. This allows for URLs to actual image binaries,
    data for on-the-fly image composition, etc. Maximum length is 1000
    characters.
-   ```snaps``` list of snaps that make up the
    product. This list will be used to for snappy based products to
    trigger tests when new Snap revisions are made available.

    The format is a list of: ```[snap_name, revision_num, locked]```

    -   ```snap_name``` is the name of the Snap in the Store
    -   ```revision``` is the revision number (integer) that was used in the base image creation
    -   ```locked``` is a boolean indicating if that revision should be locked when generating Tests or can float as snaps are updated.

In the current system, interaction is via an HTTP API and detailed below:

<table class="table">
<tr>
    <th> Action     <th>Manifest Creation
<tr>
    <td>URL         <td>https://spi.canonical.com/orgs/<org-id\>/products/manifests
<tr>
    <td>Method      <td>POST
<tr>
    <td>Data        <td>
<pre>
{
    “image_name”: "test\_image",
    “release”: “15.04”,
    “base_image_reference”: "http://foo.bar",
    “snaps”: [['snap A', 1, true],['snap B', null, false]]
}
</pre>
<tr>
    <td>Response    <td><code>201 CREATED</code> with the JSON representation of the product.
<tr>
    <td>Response    <td><code>409 CONFLICT</code> if an identical product already exists
<tr>
    <td>Example     <td>
<pre>
$ ./scripts/api_example.py -X POST conf.ini  https://spi.canonical.com/orgs/docs-org/products/manifests --data '{
  "image_name": "test_image",
  "release": "15.04",
  "base_image_reference": "http://foo.bar",
  "snaps": [["snap A", 1, true],["snap B", null, false]]
}'

HTTP 201
{
  "manifest": {
    "image_name": "test_image",
    "snaps": [
      {
        "locked": true,
        "revision": 1,
        "name": "snap A"
      },
      {
        "locked": false,
        "revision": null,
        "name": "snap B"
      }
    ],
    "base_image_reference": "http://foo.bar",
    "id": "ae02b198-bfc8-4d26-aa7b-53dbfdd38230",
    "release": "15.04",
    "created_at": "2015-08-28T20:15:15.726057"
  }
}
</pre>
</table>

<table class="table">
<tr>
    <th>Action  <th>Manifest Listing
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/products/manifests
<tr>
    <td>Method  <td>GET
<tr>
    <td>Data    <td>N/A
<tr>
    <td>Response    <td><code>200</code> a JSON list of manifests
<tr>
    <td>Example     <td>
<pre>
$ ./scripts/api_example.py conf.ini https://spi.canonical.com/orgs/docs-org/products/manifests

HTTP 200
{
  "manifests": [
    {
      "image_name": "e2e-img8c80b8b271814262add99ccf4d3deba6",
      "base_image_reference": "http://foo.com/bar",
      "snaps": [
        {
          "revision": 1,
          "name": "snap-a",
          "locked": true
        },
        {
          "revision": null,
          "name": "snap-b",
          "locked": false
        }
      ],
      "id": "9b2ff50c-5fab-4039-a5e5-1a9672d71223",
      "release": "15.04",
      "created_at": "2015-08-28T20:00:11.058236"
    }
  ]
}
</pre>
</table>

<table class="table">
<tr>
    <th>Action  <th>Manifest View
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/products/manifests/&lt;manifest-id&gt;
<tr>
    <td>Method  <td>GET
<tr>
    <td>Data    <td>N/A
<tr>
    <td>Response    <td><code>200</code> a JSON with the manifest’s details
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py conf.ini https://spi.canonical.com/orgs/docs-org/products/manifests/9b2ff50c-5fab-4039-a5e5-1a9672d71223

HTTP 200
{
  "manifest": {
    "release": "15.04",
    "base_image_reference": "http://foo.com/bar",
    "id": "9b2ff50c-5fab-4039-a5e5-1a9672d71223",
    "image_name": "e2e-img8c80b8b271814262add99ccf4d3deba6",
    "snaps": [
      {
        "locked": true,
        "name": "snap-a",
        "revision": 1
      },
      {
        "locked": false,
        "name": "snap-b",
        "revision": null
      }
    ],
    "created_at": "2015-08-28T20:00:11.058236"
  }
}
</pre>
</table>


<table class="table">
<tr>
    <th>Action  <th>Manifest Deactivation/Reactivation
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/products/manifests/&lt;manifest-id&gt;
<tr>
    <td>Method  <td>PUT
<tr>
    <td>Data    <td>
<pre>
{
  “active”: "false",  # or “true” to reactivate
}
</pre>
<tr>
    <td>Response    <td><code>200 OK</code>
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X PUT conf.ini \
https://spi.canonical.com/orgs/docs-org/products/manifests/9b2ff50c-5fab-4039-a5e5-1a9672d71223

HTTP 200
{
  "ok": true
}
</table>

<table class="table">
<tr>
    <th>Action  <th>Manifest Deletion
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/products/manifests/&lt;manifest-id&gt;
<tr>
    <td>Method  <td>DELETE
<tr>
    <td>Data    <td>N/A
<tr>
    <td>Response    <td><code>200 OK</code>
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X DELETE conf.ini \
https://spi.canonical.com/orgs/docs-org/products/manifests/9b2ff50c-5fab-4039-a5e5-1a9672d71223

HTTP 200
{
  "ok": true
}
</table>

Test Management
---------------

Tests are defined with Test Specs. There may be multiple
Test Specs per product and they are related via the image\_name field.
TestSpecs are required for both Image and Snappy based workflows as they
form the basis for TestOpportunity creation.

Test Specs are immutable but may be deleted or deactivated to be
hidden from view. The details of the Test Spec that led to a Test
Opportunity is always persisted in the Test Result data.

To modify a Test Spec, the user should delete the existing one and
create a new one with the desired changes. That will not automatically
trigger new tests, for that see [Test
Triggering](#test-triggering).

<table class="table">
<tr>
    <th>Action  <th>Test Spec Creation
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/tests/specs
<tr>
    <td>Method  <td>POST
<tr>
    <td>Data    <td>
<pre>
{
  "platform": platform of the product
  "name": unique (within an org) name of the Test
  "image_name": name that ties this test to the product/manifest
  "test_payload": an arbitrary JSON value with all the information needed to run the test
                  (maximum size 4000 bytes), optional defaulting to null
}
</pre>
<tr>
    <td>Response    <td><code>201 CREATED</code> with a Location header pointing to the spec and the JSON representation of the spec.
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X POST conf.ini https://spi.canonical.com/orgs/docs-org/tests/specs \
--data '
{
  "platform": "amd64",
  "image_name": "imgname",
  "name": "specname32"
}'

HTTP 201
{
  "spec": {
    "image_name": "imgname",
    "id": "cd98a77f-f7da-4388-b007-0a8a3604b676",
    "name": "specname32",
    "platform": "amd64",
    "created_at": "2015-08-28T20:28:46.334756",
    "test_payload": null
  }
}
</pre>
</table>

<table class="table">
<tr>
    <th>Action  <th>Test Spec List
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/tests/specs
<tr>
    <td>Method  <td>GET
<tr>
    <td>Data    <td>N/A
<tr>
    <td>Response    <td><code>200 OK</code> with the JSON list of specs
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py conf.ini https://spi.canonical.com/orgs/docs-org/tests/specs

HTTP 200
{
  "specs": [
    {
      "created_at": "2015-08-28T20:28:46.334756",
      "id": "cd98a77f-f7da-4388-b007-0a8a3604b676",
      "test_payload": null,
      "image_name": "imgname",
      "platform": "amd64",
      "name": "specname32"
    }
  ]
}
</pre>
</table>

<table class="table">
<tr>
    <th>Action  <th>Test Spec View
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/<org-id>/tests/specs/<spec-id>
<tr>
    <td>Method  <td>GET
<tr>
    <td>Data    <td>N/A
<tr>
    <td>Response    <td><code>200 OK</code> with the JSON of a single Test Spec
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py conf.ini https://spi.canonical.com/orgs/docs-org/tests/specs/cd98a77f-f7da-4388-b007-0a8a3604b676

HTTP 200
{
  "spec": {
    "id": "cd98a77f-f7da-4388-b007-0a8a3604b676",
    "name": "specname32",
    "platform": "amd64",
    "created_at": "2015-08-28T20:28:46.334756",
    "test_payload": null,
    "image_name": "imgname"
  }
}

</pre>
</table>

<table class="table">
<tr>
    <th>Action  <th>Test Spec Deactivation/Reactivation
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/<org-id>/tests/specs/<spec-id>
<tr>
    <td>Method  <td>PUT
<tr>
    <td>Data    <td>
<pre>
{
  “active”: "false",  # or “true” to reactivate
}
</pre>
<tr>
    <td>Response    <td><code>200 OK</code>
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X PUT conf.ini \
https://spi.canonical.com/orgs/docs-org/tests/specs/cd98a77f-f7da-4388-b007-0a8a3604b676

HTTP 200
{
  "ok": true
}
</pre>
</table>

<table class="table">
<tr>
    <th>Action  <th>Test Spec Deletion
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/<org-id>/tests/specs/<spec-id>
<tr>
    <td>Method  <td>DELETE
<tr>
    <td>Data    <td>N/A
<tr>
    <td>Response    <td><code>200 OK</code>
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X DELETE conf.ini \
https://spi.canonical.com/orgs/docs-org/tests/specs/cd98a77f-f7da-4388-b007-0a8a3604b676

HTTP 200
{
  "ok": true
}</pre>
</table>

Test Triggering
---------------

### Image Submission

Test opportunities are triggered by a new image reference being
advertised by the user to the system. An Image Reference
has the following required fields:



-   ```image_name```: a unique name within an
    organization that is combined with each
    test case to produce the full set of test
    opportunities. I.e. if you have two test cases with a
    matching image_name, then two test opportunities will be created.
    Maximum length is 200 characters.
-   ```image_reference```: is opaque to the
    system; it is for your Provisioning Kit in your lab
    to interpret and use. It can be any JSON
    value. This allows for URLs to actual image binaries,
    data for on-the-fly image composition, etc. Maximum length is 1000
    characters.
-   ```image_unique_id```: also opaque to the
    system. It is recommended that this uniquely
    identifies the advertised version of the image so that the
    provisioning script can verify the image\_reference and results can
    be queried by image\_unique\_id. For example it could
    be a hash of the actual binary, or a hash of the image\_reference
    contents in the case where
    image\_reference is not a URL to a binary. Currently
    this ID is used for querying test results. Maximum length is 36
    characters.
-   ```upgrade\_snaps```: NOT IMPLEMENTED YET.


<table class="table">
<tr>
    <th>Action  <th>Image Reference Creation
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/images
<tr>
    <td>Method  <td>POST
<tr>
    <td>Data    <td>
<pre>
{
    image_reference: "http://foo.bar",
    image_unique_id: some unique_id,
    image_name: my_image,
    upgrade_snaps: []
}
</pre>
<tr>
    <td>Response    <td><code>201 CREATED</code> with a list of test opportunity IDs
<tr>
    <td>Response    <td><code>404 NOT FOUND</code> if there is no test spec for this image_name
<pre>
{
  "id": "resource-not-found",
  "message": "Couldn't find a Spec for image name 'my_image'",
  "url": null
}
</pre>
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X POST conf.ini https://spi.canonical.com/orgs/docs-org/images --data \
'{
"image_reference": "http://foo.bar",
"image_unique_id": "fooo",
"image_name": "imgname",
"upgrade_snaps": []
}'

HTTP 201
{
  'ids': ['99787105-911f-4e23-bded-b0b8253aee11']
}
</pre>
</table>

### Snap Revisions

Test opportunities may be triggered by snap revisions being
advertised to the system.

New snap revisions uploaded to the Store will be automatically
advertised to the system, however they can be re-advertised manually
afterwards, re-triggering tests with their original context.

The new revision advertised will be matched with all **Manifests** within the context **Organization**
referencing the given **Snap**, and **Test Opportunities** will be created for all **Test Specs** linked to those
**Manifests**.

![](images/image03.png)


By modelling new / specific **Manifests** (Products) under the same
**Organization** customers define the testing context they care about:



1.  **Update from a Gold Master (GM) image**:

    Product "base\_image\_reference" contains specific versions of
    product-specific snaps reflecting one generation / portion of the
    installed base. In this case, tests represent the act of updating
    products of an specific generation to the advertised snap.


2.  **Fresh Install (special case of above where GM == Ubuntu-Core)**:

    Product "base\_image\_reference" does not include
    product-specific snaps and tests represent, for instance, the act of
    installing the advertised snap on top of a vanilla ubuntu-core
    image.


From the [Agent & Provisioning Kit](#agent-operation) point of view, testing contexts will
be homogeneously defined via Test Opportunity base\_image\_reference and extra\_snaps.

Snap revision advertisement request requires the following
fields:

-   **name** The name of the snap package. Maximum length is 200 characters.
-   **revision** The new snap's revision. An integer number.



Snap revision advertisement will return a list of test opportunity IDs and image unique IDs that can be used to [query for results](#querying-test-results) later.

<table class="table">
<tr>
    <th>Action  <th>Snap Revision Triggering
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org-id&gt;/snap-revisions
<tr>
    <td>Method  <td>POST
<tr>
    <td>Data    <td>
<pre>
{
    "name": name of snap package
    "revision": revision number in the Store (integer)
}
</pre>
<tr>
    <td>Response    <td><code>201 CREATED</code> with a list of test opportunity IDs and image unique IDs
<tr>
    <td>Response    <td><code>404 NOT FOUND</code> if there is no matched Manifests
<pre>
{
  "id": "resource-not-found",
  "message": "",
  "url": null
}
</pre>
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py -X POST conf.ini https://spi.canonical.com/orgs/docs-org/snap-revisions \
--data '
> {
>   "name": "hello.world",
>   "revision": 123
> }'

HTTP 201
{
'ids': [{'test_opportunity_id': '7b6e755d-28c9-4663-bcfc-45c36eee6eb3',
      	'image_unique_id': '5f88cc90-487b-4587-a633-b36e44b72b99'}]}
</pre>
</table>


Querying Test Progress
----------------------



As a test opportunity progresses from being triggered to deliver a
result it flows through a series of expected events. This is the normal
expected event flow for a test opportunity:

![](images/image00.png)

Note, if there are multiple agent-groups or a test run times out
and is retried, there may be
multiple PICKED\_BY\_AGENT, RESULT\_POSTED,
RESULT\_STORED events for a given test opportunity.

An API is provided to retrieve event logs for a
given test opportunity . Clients may use
this to track test opportunity progress and to debug
cases where results have not been generated in a timely
manner.  The Test Progress API can be used to query
progress for individual test\_opportunity\_ids or in
aggregate by a variety of other fields as detailed below.

<table class="table">
<tr>
    <th>Action  <th>Result Query
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org_id&gt;/tests/events
<tr>
    <td>Method  <td>GET
<tr>
    <td>Query Params    <td>The following are optional query filters:
<ul>
    <li>test_opportuniy_id
    <li>spec_name
    <li>test_spec_id
    <li>image_name
    <li>image_unique_id
    <li>start_date
    <li>end_date
    <li>with_event
    <li>without_event
</ul>
<p>All parameters do exact matching, except start_date and end_date that work creating a date range (both inclusives). Dates are in ISO8601 format.

<p>Using with_event=EVENT_TYPE will return only logs that contain that event type.
For example with_event=RESULT_POSTED returns only logs for tests for which a result has been posted back. Using without_event=EVENT_TYPE will return only logs that do not contain that event type. with_event and without_event cannot be used in the same query.

<p>For example, a large number of results that match without_event=PICKED_BY_AGENT could indicate either an under-resourced lab (all agents are busy) or that the agents have lost communication with Canonical's servers. The combination of with_event=PICKED_BY_AGENT and without_event=RESULT_POSTED can be used to track down slow or stuck test runs, and the agent_id fields of the PICKED_BY_AGENT event can be used to identify the problem agent in the lab.
<tr>
    <td>Response    <td><code>200 OK</code> with JSON list of matching progress event log documents
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py conf.ini \
http://spi.canonical.com/orgs/e2e-org/tests/events?test_opportunity_id=d012edab-9dd1-4950-bdf6-21b94a956791
HTTP 200
{
  "event_logs": [
    {
      "organization_id": "e2e-org",
      "image_name": "e2e-imgf151a8cd9081404db54ed1d8dab1e140",
      "queued_at": "2015-09-17T12:13:03.288000+00:00",
      "updated_at": "2015-09-17T12:13:06.799000+00:00",
      "image_unique_id": "3a9cf076-1c27-4b1c-8080-15cb2d24bea7",
      "events_seen": [
        "QUEUED_FOR_AGENTS",
        "PICKED_BY_AGENT",
        "RESULT_POSTED",
        "RESULT_STORED"
      ],
      "events": [
        {
          "event_type": "QUEUED_FOR_AGENTS",
          "timestamp": "2015-09-17T12:13:03.288000+00:00"
        },
        {
          "event_type": "PICKED_BY_AGENT",
          "agent_group": "e2e-group",
          "agent_id": "e2e-id",
          "timestamp": "2015-09-17T12:13:06.291000+00:00",
          "platform": "amd64"
        },
        {
          "result_id": "533050b7-054e-4a72-bf63-16db864451ea",
          "test_status": "PASSED",
          "agent_group": "e2e-group",
          "event_type": "RESULT_POSTED",
          "timestamp": "2015-09-17T12:13:06.734000+00:00",
          "agent_id": "e2e-id",
          "platform": "amd64"
        },
        {
          "event_type": "RESULT_STORED",
          "result_id": "533050b7-054e-4a72-bf63-16db864451ea",
          "test_status": "PASSED",
          "timestamp": "2015-09-17T12:13:06.745000+00:00"
        }
      ],
      "test_opportunity_id": "d012edab-9dd1-4950-bdf6-21b94a956791",
      "spec_name": "e2e-spec_13691451",
      "test_opportunity": {

(...)
</pre>
</table>


### Data in Progress Logs


There are a few key fields in the test event logs JSON of
note:

-   **test\_opportunity\_id**: the id of the
    test opportunity the logs are about.
-   **queued\_at**: the timestamp of when the
    test opportunity was queued for agent dispatch.
-   **updated\_at**: a timestamp of the last
    time the logs were updated with an event.
-   **test\_opportunty**: the full test
    opportunity object that was sent to agents.
-   **events**: an array with the actual
    progress event objects, each object with the fields:

    -   **event\_type**: the event type name (one
    of QUEUED\_FOR\_AGENTS, etc)
    -   **timestamp**: timestamp of the event
    -   further event type specific fields, see the next table

-   **spec\_name**: the name of the TestSpec that originated the Test Opportunity that generated this result.
-   **image\_name**: the image name from the originating TestSpec

<table class="table">
<tr>
    <th>Event Type      <th>Event type specific fields
<tr>
    <td>QUEUED_FOR_AGENTS                <td>none
<tr>
    <td>PICKED_BY_AGENT                <td>platform, agent_group, agent_id
<tr>
    <td>RESULT_POSTED                <td>platform, agent_group, agent_id, result_id, test_status
<tr>
    <td>RESULT_STORED                <td>result_id, test_status
</table>


Querying Test Results
---------------------


Querying for results is done with the following API:

<table class="table">
<tr>
    <th>Action  <th>Result Query
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org_id&gt;/results
<tr>
    <td>Method  <td>GET
<tr>
    <td>Query Params    <td>The following are optional query filters:
<ul>
<li>image_name
<li>spec_name
<li>image_unique_id
<li>start_date
<li>end_date
</ul>

<p>All parameters do exact matching, except start_date and end_date that work creating a date range (both inclusives). Dates are in ISO8601 format.

<tr>
    <td>Response    <td><code>200 OK</code> with JSON list of matching Test Results
<tr>
    <td>Example    <td>
<pre>
./scripts/api_example.py conf.ini \
'https://spi.canonical.com/orgs/docs-org/results?start_date=20150826T00%3A00%3A00&end_date=20150826T04%3A00%3A00'

HTTP 200
{
  "test_results": [
    {
      "test_spec": {
        "name": "e2e-spec_70600150",
        "platform": "amd64",
        "image_name": "e2e-img",
        "id": "9e12839e-d9c1-4266-be33-27b5174d396f",
        "created_at": "2015-08-26T00:00:20.297477",
        "test_payload": "{'foo': 'bar'}"
      },
      "test_status": "PASSED",
      "platform": "amd64",
      "organization_id": "docs-org",
      "image_name": "e2e-img",
      "extra_snaps": [],
      "image_reference": "http://foo.com/bar",
      "agent": {
        "timing": {
          "setup_start": "2015-08-26T00:00:22.651078",

[...]
</pre>
</table>


### A note about dates

When querying for results you can specify start\_date and/or
end\_date. We formally support [ISO 8601
format](https://www.google.com/url?q=https://en.wikipedia.org/wiki/ISO_8601&sa=D&usg=AFQjCNG1YKr9awDhJrozB9Cs0KgC5Ps-6g) for
these dates, but we’re also more flexible, allowing you to use different
formats:

-   2015-09-14T17:51:31+00:00
-   2015-09-14
-   2015/09/14
-   20150914
-   20150914 17:51:31
-   20150914 175131 Z



### Test Result Data



There are a few key fields in the test result JSON of note:

<!-- FIX link to actors-and-results -->

-   **test\_status**: result status based on
    the exit codes of provisioning and test scripts. See [below.](#actors-and-results)
-   **result\_payload**: the full JSON that was written to **spi\_test\_result.json** by the test script.
-   **spec\_name**: the name of the TestSpec that originated the Test Opportunity that generated this
    result.
-   **image\_name**: the image name from the originating TestSpec

*All other fields provide further data about the inputs that created originating Test Opportunity (e.g. "image" and "test\_spec"), and are subject to change.*

### Deprecated Query API



The following is a deprecated (as of 2015-08-07 release) endpoint
for querying results with the ```image_unique_id``` used when [the image reference was published](#image-submission).



<table class="table">
<tr>
    <th>Action  <th>Result Query
<tr>
    <td>URL     <td>https://spi.canonical.com/orgs/&lt;org_id&gt;/results/&lt;image_unique_id&gt;
<tr>
    <td>Method  <td>GET
<tr>
    <td>Response    <td><code>200 OK</code> with JSON list of Test Results
<tr>
    <td>Example    <td>
<pre>
$ ./scripts/api_example.py conf.ini https://spi.canonical.com/orgs/docs-org/results/9e98346a-07e7-45a9-8105-b25b50bd6fa1

HTTP 200
{
  "test_results": [
    {
      "image_unique_id": "9e98346a-07e7-45a9-8105-b25b50bd6fa1",
      "test_status": "PASSED",
      "agent-id": "e2e-id",
      "id": "37383ba5-ff0b-4735-917a-ac4a4b8939c9",
      "result_payload": {
        "foo": "bar"
      },
      "image_reference": "http://foo.com/bar",
      "organization_id": "docs-org",
      "test_payload": "{'foo': 'bar'}",
      "created_at": "2015-08-26T03:00:11.276000+00:00",
[...]
</pre>
</table>
