import logging
import os
import sys
import requests
import subprocess


PACKAGE_URL = "https://search.apps.ubuntu.com/api/v1/package/{}"


def get_prov_kit_logger():
    """Configure logging and return the provkit logger."""
    logging.basicConfig(format="%(asctime)s %(levelname)s %(name)s: %(message)s",
                        stream=sys.stderr, level=logging.INFO)
    logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.ERROR)
    return logging.getLogger("provkit.{}".format(os.path.basename(sys.argv[0])))


log = get_prov_kit_logger()


def device_sshopts():
    opts = ['-o', 'StrictHostKeyChecking=no', '-o', 'UserKnownHostsFile=/dev/null']

    class sshopts:
        ip = os.getenv('RPI2_IP', 'webdm.local')
        port = int(os.getenv('RPI2_SSH_PORT', 22))
        target = 'ubuntu@{}'.format(ip)
        password = "ubuntu"
        ssh = opts + ['-q', '-p', str(port), target]
        scp = opts + ['-q', '-P', str(port)]
    return sshopts


def get_snap_info(arch, release, channel, snap_name, snap_seq, bail_on_seq_mismatch=True):
    """ Fetch package info from Click Package Index (CPI) and return the download URL
    and a boolean is_oem if this is an OEM type package. """

    headers = {
        'X-Ubuntu-Architecture': arch,
        'X-Ubuntu-Release': release
    }

    url = PACKAGE_URL.format(snap_name)
    log.info('Requesting package info for snap: %s, url: %s, headers: %r', snap_name, url, headers)

    res = requests.get(url, headers=headers)
    if res.status_code != 200:
        log.error('Cannot find package info for snap: %s, status_code: %d',
                  snap_name, res.status_code)
        sys.exit(1)
    info = res.json()
    if info['sequence'] != snap_seq:
        log.error("cannot get hold of {} sequence {}, it's not the current latest: {}"
                  .format(snap_name, snap_seq, info['sequence']))
        sys.exit(1)

    url = info['anon_download_url']
    is_oem = info['content'] == 'oem'
    return url, is_oem


def download_snap(url, output_filename):
    # TODO: consider keeping a cache of downloads
    log.info("Downloading: {}".format(url))
    subprocess.check_call(['curl', '-o', output_filename, url])


def compose_image(arch, release, channel, base_snaps, image_filename):
    """ Gather snap info, downlaod snaps, and use ubuntu-device-flash to compose an image"""
    # gather snap info
    oem = None
    installs = []
    for s in base_snaps:
        if s['name'] == 'ubuntu-core':  # XXX: remove once we have all-snaps
            log.warn('Skipping ubuntu-core snap because we are still in hybrid system-image world')
            core = s
            continue
        s['url'], s['is_oem'] = get_snap_info(arch, release, channel, s['name'], s['sequence'])
        s['fname'] = '{}_s{}.snap'.format(s['name'], s['sequence'])
        if s['is_oem']:
            oem = s['fname']
        else:
            installs += ['--install', s['fname']]

    if core is not None:
        base_snaps.remove(core)
    if oem is None:
        log.error('OEM snap is required to be specified in the Product, but none found')
        sys.exit(1)
    # log.info('base_snaps: %r', base_snaps)
    log.info('oem: %s, installs: %r', oem, installs)

    # downlaod snaps - TODO: parallelize this
    for s in base_snaps:
        download_snap(s['url'], s['fname'])

    # TODO: consider writing the udf cmd or the list of snaps/seqs out next to the image

    sys_image_release = 'rolling'  # XXX: ditch when all-snaps
    sys_image_chan = 'edge'  # XXX: ditch when all-snaps
    udf_args = ['sudo', 'ubuntu-device-flash', 'core', sys_image_release,
                '--developer-mode',
                '--device', 'raspi2_armhf',
                '-o', image_filename,
                '--channel', sys_image_chan,
                '--oem', oem
                ] + installs

    log.info('Composing image with: %s', ' '.join(udf_args))
    log.warn('Note: you may be required to enter your password for sudo. This is a result of the '
             'current ubuntu-device-flash requiring root access.')
    subprocess.check_call(udf_args)
