#!/usr/bin/env python3
"""
SPI Api_example - example wrapper for making oauth-signed calls to SPI API endpoints

Usage:
  api_example.py [options] <configfile> <url>

Options:
  -h --help                 Show this screen
  -d, --data data|@<file>   POST Data
  -X method                 HTTP Method
  -v                        Verbose debug output
"""

import json
import logging
import os
import sys

from configparser import ConfigParser
import requests_oauthlib

logging.basicConfig(format='%(asctime)s %(levelname)-5.5s %(name)s %(message)s')
logger = logging.getLogger(__name__)

try:
    import docopt
except:
    print('This script must be run in the virtualenv, via:')
    print('. env/bin/activate')
    print('./scripts/api_example.py')
    sys.exit(0)


class ApiExample(object):

    def __init__(self, arguments):
        self.config = ConfigParser()
        self.confpath = arguments['<configfile>']
        self.method = arguments.get('-X')
        self.url = arguments.get('<url>')
        pd = arguments.get('--data')
        if pd and pd.startswith('@'):
            # Filename specified, replace pd with data from the file
            with open(os.path.expanduser(pd[1:])) as datafile:
                self.post_data = json.load(datafile)
        else:
            self.post_data = json.loads(pd) if pd else None

        if arguments.get('-v'):
            logger.setLevel(logging.DEBUG)
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True
            requests_oauthlib_log = logging.getLogger("requests_oauthlib")
            requests_oauthlib_log.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

        # Load config
        if self.confpath is not None:
            self.config.read(self.confpath)

        if 'auth' in self.config:
            # Here's the substance - we create a requests session
            # with our oauth credentials (read from config file)
            self.session = requests_oauthlib.OAuth1Session(
                self.config.get('auth', 'consumer_key'),
                client_secret=self.config.get('auth', 'consumer_secret'),
                resource_owner_key=self.config.get('auth', 'token_key'),
                resource_owner_secret=self.config.get('auth', 'token_secret'),
            )
        else:
            logger.warning("[auth] config section not found in %s", self.confpath)
            sys.exit(1)

    def run(self):
        if self.method is None:
            if self.post_data is not None:
                self.method = 'POST'
            else:
                self.method = 'GET'
        logger.debug('%s %s', self.method, self.url)
        logger.debug('with data: %s', self.post_data)

        if self.method == 'POST':
            res = self.session.post(self.url, json=self.post_data)
        elif self.method == 'PUT':
            res = self.session.put(self.url, json=self.post_data)
        elif self.method == 'DELETE':
            res = self.session.delete(self.url)
        else:
            res = self.session.get(self.url)

        print('HTTP', res.status_code)
        try:
            print(json.dumps(res.json(), indent=2))
        except ValueError:
            # response is not json
            print(res.text)
            sys.exit(1)


if __name__ == '__main__':
    arguments = docopt.docopt(__doc__, version='SPI Api_example v1')
    api = ApiExample(arguments)
    api.run()
