#!/usr/bin/env python3

# © Canonical 2015
# Licensed under AGPLv3

"""
Tanuki End to End Test

Usage:
  e2e-noauth.py [-i] <configfile>
  e2e-noauth.py --default-config

Options:
  -h --help             Show this screen
  --default-config      Show default configuration
  --server              MessageHub URL override
  --specmgr             SpecManager URL override
  -i                    Interactive
"""


# Run this script without [auth] in the config to check for non-authed access
# Run this script with [auth] in the config but wrong org-id to check for unrestricted access

import logging
import subprocess
import sys

import docopt

from e2e import E2e, logger, nuitka

logger.setLevel(logging.ERROR)


class E2eNoAuth(E2e):

    testopp_ids = ['7e92267d-d212-4966-80e2-8a19c4ad906e']
    spec_id = 'DEADBEEF'

    def handle_error(self, result, when):
        """All things should 401 all the time."""
        if 200 < result.status_code < 299:
            logger.error(
                'Unexpected successful code %d when %s at %s: %s',
                result.status_code,
                when,
                result.url,
                result.text)
            exit(1)
        else:
            result.raise_for_status()

    def annotate(self, *a, **kw):
        """Ignore annotations."""
        pass

    def run(self):
        self.annotate('Starting.')
        logger.info('Server : %s', self.server)

        if 'auth' in self.config:
            errmsg = '%s endpoint accessed with wrong credentials'
        else:
            errmsg = '%s endpoint did not require auth'

        for f in [
                self.cleanup_manifests,
                self.create_manifest,
                self.cleanup_specs,
                self.create_spec,
                self.submit_image,
                self.submit_snap_revision,
                self.check_progress,
                self.check_results,
                self.delete_spec]:
            try:
                f()
                logger.error(errmsg, f.__name__)
                self.exit_code = 1
            except:  # We ignore all failure
                pass

        # Also check agent fails the right way
        if nuitka:
            data = subprocess.check_output(
                './agent.py -v --register {}'.format(self.confpath),
                stderr=subprocess.STDOUT,
                shell=True)
        else:
            data = subprocess.check_output(
                '{} agent.py -v --register {}'.format(sys.executable, self.confpath),
                stderr=subprocess.STDOUT,
                shell=True)

        if 'Error 401' not in data.decode('utf-8'):
            logger.error(errmsg, 'agent registration')
            self.exit_code = 1


if __name__ == '__main__':
    arguments = docopt.docopt(__doc__, version='Tanuki E2e Proto 1')
    e2e = E2eNoAuth(arguments)
    e2e.run()
    exit(e2e.exit_code)
