#!/usr/bin/env python3

# © Canonical 2015
# Licensed under AGPLv3

"""
Tanuki End to End Test

Usage:
  e2e-image.py [-i] <configfile>
  e2e-image.py --default-config

Options:
  -h --help             Show this screen
  --default-config      Show default configuration
  --server              MessageHub URL override
  --specmgr             SpecManager URL override
  -i                    Interactive
"""

import docopt
import time

from e2e import E2e, logger, GRY


class E2eUpdate(E2e):

    def run(self):
        self.annotate('Starting.')
        logger.info('Server : %s', self.server)
        self.cleanup_products()
        self.annotate('As Developer or QA, Create Product via Spec Manager')
        self.create_product()
        self.submit_gm()
        self.annotate('As Developer or QA, Create Test Spec via Spec Manager')
        self.create_spec()
        self.prompt(
            'As Lab Owner, Run agent first to register with Message Hub')
        self.run_agent(reg_only=True)
        self.pause()
        self.submit_snap()
        self.annotate('Sleeping 1s to wait for propagation to hub', color=GRY)
        time.sleep(1)
        self.annotate('As Lab Owner, Running agent')
        self.annotate(
            'Should pick up at least 1 Test Opportunity,'
            ' provision and run tests', color=GRY)
        self.annotate('Note this can also run continously in daemon mode',
                      color=GRY)
        self.pause()
        self.run_agent()
        self.prompt('Completed agent run')
        self.annotate('As Developer or QA, Checking result progress via Result Enumerator')
        self.check_progress()
        self.annotate('As Developer or QA, Checking final results via Result Enumerator')
        results = self.check_results()
        if self.exit_code == 0:  # No need to verify it anything failed
            self.verify_results(results)
        self.cleanup_products()

    def verify_results(self, results):
        for r in results['test_results']:
            if r['test_opportunity_id'] in self.testopp_ids:
                # Pop random data
                for k1, k2 in [
                        ['test_spec', 'created_at'],
                        ['product', 'created_at'],
                        ['agent', 'timing']]:
                    try:
                        r[k1].pop(k2)
                    except KeyError:
                        logger.error('Missing key in result: r[{}][{}]'.format(k1, k2))
                        self.exit_code = 1

                for k in "image_unique_id", "base_image_id":
                    i = r.pop(k)
                    if i not in self.image_unique_ids:
                        logger.error("%s is %s, not in %r", k, i, self.image_unique_ids)
                r["update_snaps"] = sorted(r["update_snaps"])

                expected_data = {
                    "product_id": self.product_id,
                    "platform": "amd64",
                    "base_channel": "stable",
                    "base_image_reference": None,
                    "agent_id": "e2e-id",
                    "update_snaps": sorted([
                        ["webdm.canonical", 1],
                        ["ubuntu-core.canonical", 1],
                        ["test-snap.%s" % self.namespace, 2],
                        ["pi2.canonical", 1]
                    ]),
                    "spec_name": self.spec_name,
                    "image_reference": None,
                    "product": {
                        "active": True,
                        "primary_snap_name": "test-snap.%s" % self.namespace,
                        "name": "test-product.%s" % self.namespace,
                        "release": "rolling-core",
                        "snaps": [
                            "ubuntu-core.canonical",
                            "pi2.canonical",
                            "webdm.canonical",
                            "test-snap.%s" % self.namespace
                        ],
                        "id": self.product_id
                    },
                    "agent": {
                        "id": "e2e-id"
                    },
                    "test_payload": "{'foo': 'bar'}",
                    "test_status": "PASSED",
                    "result_payload": {
                        "foo": "bar"
                    },
                    "base_snaps": [
                        "ubuntu-core.canonical",
                        "pi2.canonical",
                        "webdm.canonical",
                        "test-snap.%s" % self.namespace
                    ],
                    "primary_snap_name": "test-snap.%s" % self.namespace,
                    "type": "update",
                    "test_spec": {
                        "active": True,
                        "product_id": self.product_id,
                        "channel_combos": [{"update": "edge", "base": "stable"}],
                        "name": self.spec_name,
                        "test_payload": "{'foo': 'bar'}",
                        "platform": "amd64",
                        "id": self.spec_id
                    },
                    "product_name": "test-product.%s" % self.namespace,
                    "update_channel": "edge"
                    }

                for k, v in expected_data.items():
                    try:
                        if r[k] != v:
                            self.exit_code = 1
                            logger.error('expected result[{}] == {},'
                                         ' got {} instead'.format(k, v, r[k]))
                    except KeyError:
                        logger.error('Missing key in result: {}'.format(k))
                        self.exit_code = 1

if __name__ == '__main__':
    arguments = docopt.docopt(__doc__, version='Tanuki E2e Proto 1')
    e2e = E2eUpdate(arguments)
    e2e.run()
    exit(e2e.exit_code)
