#!/usr/bin/env python3
"""
SPI Tutorial

Usage:
  tutorial.py [-i] [<configfile>]
  tutorial.py --default-config

Options:
  -h --help             Show this screen
  --default-config      Show default configuration
  --server              MessageHub URL override
  --specmgr             SpecManager URL override
  -i                    Interactive
"""

from configparser import ConfigParser

import subprocess
import sys
import time

try:
    import docopt
except:
    print('This script must be run in the virtualenv, via:')
    print('. env/bin/activate')
    print('./scripts/tutorial.py')
    sys.exit(0)

from e2e import E2e, logger, GRY


class E2eImage(E2e):

    def run(self):
        self.annotate('Welcome to the SPI Tutorial!')
        self.annotate('----------------------------')

        self.annotate('We are going to walkthrough an image-based workflow,')
        self.annotate('performing a simple test case.')
        self.annotate('For the purposes of this tutorial we will skip any')
        self.annotate('actual device provisioning.')
        self.pause()

        if 'auth' in self.config:  # Already authenticated
            self.annotate('I see you are already authenticated.')
            self.annotate('Skipping authentication section.')
        else:
            self.annotate('First, we need to authenticate into the system.')
            self.annotate('Create an account at https://login.ubuntu.com if you don\'t have one.')
            self.pause()
            self.annotate('Authenticating...')
            self.annotate('./agent.py --interactive-login scripts/tutorial.ini')
            subprocess.check_call(['./agent.py', '--interactive-login', 'scripts/tutorial.ini'])

            self.annotate('Your credentials are stored in the scripts/tutorial.ini '
                          'confguration file')
            self.pause()
            self._load_config()

        if not self.organization_id:
            self.annotate('As part of the onboarding process, you have to associate your account '
                          'ID/consumer_key to an organization. Your ID/consumer_key is: %s .'
                          % self.config.get('auth', 'consumer_key'))
            self.annotate('Contact your SPI support representative to have this ID added to your '
                          'existing organization or create a new organization with a name of your '
                          'choosing.')
            org_id = input('What is your organization\'s ID? ')
            self.update_orgid(org_id)
            self.pause()

            self.annotate('The organization_id has to be consistent all accross your '
                          'usage of SPI.')
            self.annotate('This ends the authentication part of the tutorial.')
            self.pause()
        self.cleanup_specs()

        self.annotate('Now, we\'ll put on our QA hat and create a TestSpec')
        self.annotate('that describes Tests to be run for a given "image_name".')
        self.pause()
        self.annotate('The "test_payload" field is the meat of the TestSpec and will be available')
        self.annotate('to the setup and test stages of the Lab\'s Provisioning Kit')
        self.pause()
        self.annotate('SPI treats test_payload as arbitrary data.')
        self.annotate('It could be the URL of a tests tarball, JSON describing test suites, etc.')
        self.annotate('as long as the Test Spec author and the Provisioning Kit scripts')
        self.annotate('agree on the format.')
        self.pause()
        self.annotate('Here\'s an example Spec creation:')
        self.annotate('NOTE: all requests to SPI APIs will be oauth-signed,', GRY)
        self.annotate('See the User Docs for details on authenticating requests.', GRY)
        self.create_spec()

        self.annotate('Now we will setup a simple "Lab", running an Agent locally.')
        self.annotate('We are using the following agent config:')
        # XXX dump the raw file instead?
        logger.info({k: v for (k, v) in self.config.items('config')})

        self.prompt('We run the agent once to ensure we can register with SPI')
        self.annotate('./agent.py --register scripts/tutorial.ini')
        self.run_agent(reg_only=True)
        self.pause()
        self.annotate('At this point we have our Tests configured and ')
        self.annotate('an Agent ready to accept Test Opportunities.')
        self.annotate('The next step is to advertise an Image to trigger those tests.')
        self.pause()
        self.annotate('We use the same "image_name" we used before')
        self.annotate('so that the Test Spec we created will be matched.')
        self.pause()
        self.annotate('The "image_reference" is what will be handed over')
        self.annotate('to the Provisioning Kit "provcommand" in the Lab.')
        self.annotate('SPI treats it as arbitrary data, but typically this will be either')
        self.annotate('a URL to an image binary or other info for the provcommand to')
        self.annotate('install the image and ready the device for test run.')
        self.pause()
        self.annotate('"image_unique_id" is simply a hash or other identifier')
        self.annotate('unique to this submission')
        self.submit_image()
        self.annotate('Sleeping 1s to wait for the Test Opportunities to')
        self.annotate('propagate through the system')
        time.sleep(1)
        self.annotate('Now, back as the Lab Owner, we can run the agent')
        self.annotate('which should pick up at least 1 Test Opportunity, provision and run tests')
        self.annotate('Note this can also run continously in daemon mode', color=GRY)
        self.pause()
        self.run_agent()
        self.annotate('Completed agent run.')
        self.annotate('The Agent did its work and dispatched a Test Result to the SPI system.')

        self.annotate('Any participant of the organization can now query those results.')
        self.pause()
        self.check_results()
        self.annotate('Finally we\'ll cleanup, which would not be required in normal usage.')
        self.delete_spec(verbose=False)
        self.pause()
        self.annotate('All done!')
        self.annotate('Please refer to the SPI User Docs for further information')
        self.annotate('on Agent operation and API details.')

    def update_orgid(self, org_id):
        """Update this E2E's org_id."""
        cf = ConfigParser()
        cf.read(self.confpath)
        cf['config']['organization_id'] = org_id
        with open(self.confpath, 'w') as fd:
            cf.write(fd)
        self._load_config()

if __name__ == '__main__':
    arguments = docopt.docopt(__doc__, version='SPI Tutorial v1')
    arguments['-i'] = True
    arguments['<configfile>'] = 'scripts/tutorial.ini'
    e2e = E2eImage(arguments)
    e2e.run()
    exit(e2e.exit_code)
