# © Canonical 2015
# Licensed under AGPLv3

import builtins
from collections import defaultdict
from configparser import ConfigParser
import datetime
import getpass
import json
import logging
import os
import shutil
import tempfile
import time
import urllib.parse
from unittest.mock import MagicMock, call, ANY

import pytest
import requests
import requests_oauthlib

opts = {
    '<configfile>': 'conf',
    '-v': True,
}
# Default all other options to False
opts = defaultdict(lambda: False, opts)


@pytest.fixture(scope='function')
def agent(request):
    import agent
    conf = ConfigParser()
    conf['config'] = agent.DEFAULT_CONFIG
    cf = tempfile.NamedTemporaryFile(delete=False, encoding='utf-8', mode='w')
    conf.write(cf)
    cf.close()
    opts['<configfile>'] = cf.name
    request.addfinalizer(lambda p=cf.name: os.unlink(p))
    return agent.Agent(opts)


class Fake:
    pass

    def json(self):
        return self._json

    def text(self):
        return self._text


epoch = datetime.datetime(1970, 1, 1)


class FakeDate:

    def utcnow():
        return epoch


def test_register_url(agent):
    """Test register url looks right."""
    parts = urllib.parse.urlparse(agent.register_url)
    assert parts.path == "/labs/agent-queue"


def test_register_failure(agent, monkeypatch):
    """Test registration failure."""
    agent.register_url = 'http://example.com/'

    def mockpost(*a, **kw):
        x = Fake()
        x.status_code = 503
        x.text = 'ERROR'
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    result = agent.register()
    assert result is False


def test_register_bad_response(agent, monkeypatch, caplog):
    """Test registration getting bad data."""
    agent.register_url = 'http://example.com/'

    def mockpost(*a, **kw):
        assert 'agent_id' in kw['json']
        x = Fake()
        x.status_code = 200
        x._json = {'foo': 'bar'}
        return x

    with caplog.atLevel(logging.ERROR):
        monkeypatch.setattr(agent.session, 'post', mockpost)
        result = agent.register()
        assert result is False
        assert "Malformed registration response" in caplog.text()


def test_register_success(agent, monkeypatch, caplog):
    """Test registration success."""
    agent.register_url = 'http://example.com/'

    def mockpost(*a, **kw):
        assert 'agent_id' in kw['json']
        x = Fake()
        x.status_code = 200
        x.text = 'OK'
        x._json = {'packages': ['pkg1', 'pkg2'],
                   '_links': {'dequeue': {'href': 'foobar'}}}
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    with caplog.atLevel(logging.INFO):
        result = agent.register()
    assert result is True
    assert agent.queue_url == 'https://spi.canonical.com/foobar'  # NOQA
    assert "Available products: ['pkg1', 'pkg2']" in caplog.text()


def test_register_connfail(agent, monkeypatch):
    """Test registration failure to connect."""
    agent.register_url = 'http://example.com/'

    mock = MagicMock(side_effect=requests.exceptions.ConnectionError)
    monkeypatch.setattr(agent.session, 'post', mock)
    result = agent.register()
    assert result is False


def test_poll_badcall(agent):
    """Called poll unregistered."""
    with pytest.raises(SystemExit):
        agent.poll()


def test_poll_empty_200(agent, monkeypatch):
    """Empty 200 means no tests available."""
    agent.queue_url = 'http://example.com'

    def mockpost(*a, **kw):
        x = Fake()
        x.status_code = 200
        x._json = {}
        x._text = 'ERR'
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    r = agent.poll()
    assert r is False


def test_poll_any_err(agent, monkeypatch):
    """Any failure on poll."""
    agent.queue_url = 'http://example.com'

    def mockpost(*a, **kw):
        x = Fake()
        x.status_code = 300
        x._json = {'foo': 'bar'}
        x._text = 'ERR'
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    r = agent.poll()
    assert r is False


def test_poll_bad_response(agent, monkeypatch, caplog):
    """Poll receives bad data."""
    agent.queue_url = 'http://example.com'

    def mockpost(*a, **kw):
        x = Fake()
        x.status_code = 200
        x._json = {'foo': 'bar'}
        x._text = 'ERR'
        return x

    with caplog.atLevel(logging.ERROR):
        monkeypatch.setattr(agent.session, 'post', mockpost)
        r = agent.poll()
        assert r is False
        assert "Malformed poll response" in caplog.text()


def test_poll_success(agent, monkeypatch):
    """Poll succeeds."""
    agent.queue_url = 'http://example.com'

    def mockpost(*a, **kw):
        x = Fake()
        x.status_code = 200
        x._json = {
            'foo': 'bar',
            '_links': {
                'result': {'href': 'bat'}
            }
        }
        x._text = 'ERR'
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    r = agent.poll()
    assert r == {'foo': 'bar'}
    assert agent.testresult_url == 'https://spi.canonical.com/bat'  # NOQA


def test_poll_malformed_response(agent, monkeypatch):
    """Missing a result link."""
    agent.queue_url = 'http://example.com'

    def mockpost(*a, **kw):
        x = Fake()
        x.status_code = 200
        x._json = {
            'foo': 'bar',
        }
        x._text = 'ERR'
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    r = agent.poll()
    assert r is False


def test_dispatch_logging_joined(agent):
    """Test if stdout and err are logged jointly."""
    out_log = tempfile.NamedTemporaryFile()
    r = agent.dispatch(
        'fname',
        '/bin/echo foo ; /bin/echo bar 1>&2',
        out_log.name,
        '.')
    out_log.seek(0)
    assert out_log.read() == b'foo\nbar fname\n'
    assert r is True


def test_dispatch_command_failure(agent, caplog):
    """Dispatch to a failing command."""
    with caplog.atLevel(logging.INFO):
        r = agent.dispatch('foo', '/bin/false', '/dev/null', '.')
        assert 'Running "/bin/false' in caplog.text()
        assert r is False


def test_post_result_anyerr(agent, monkeypatch):
    """post_result random failure."""
    agent.queue_url = 'http://example.com'
    agent.testresult_url = 'http://example.com/42'

    def mockpost(*a, **kw):
        assert a[0] == 'http://example.com/42'
        x = Fake()
        x.status_code = 300
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    r = agent.post_result('foo')
    assert r is False


def test_post_result_badcall(agent, caplog):
    """post_result called before polling succeeded."""
    agent.queue_url = 'http://example.com'
    agent.testresult_url = None
    with caplog.atLevel(logging.ERROR):
        r = agent.post_result('foo')
        assert 'post_result called without testresult_url set' in caplog.text()
    assert r is False


def test_post_result_success(agent, monkeypatch):
    """post_result succeeds."""
    agent.queue_url = 'http://example.com'
    agent.testresult_url = 'http://example.com/42'

    def mockpost(*a, **kw):
        assert a[0] == 'http://example.com/42'
        x = Fake()
        x.status_code = 200
        return x

    monkeypatch.setattr(agent.session, 'post', mockpost)
    r = agent.post_result({})
    assert r is True


def test_post_result_fail(agent, monkeypatch, caplog):
    """post_result connection failure."""
    agent.queue_url = 'http://example.com'
    agent.testresult_url = 'http://example.com'

    mock = MagicMock(side_effect=requests.exceptions.ConnectionError)
    monkeypatch.setattr(agent.session, 'post', mock)
    with caplog.atLevel(logging.ERROR):
        r = agent.post_result('foo')
        assert 'Exception:' in caplog.text()
        mock.assert_called_with(
            'http://example.com',
            json='foo', timeout=10.0
        )
        assert r is False


def test_process_test_success(agent, monkeypatch):
    """Provisioning and test phases succeed."""
    test_opp = {
        "id": "42",
        "product": {"primary_snap_name": "foo"},
        "image_reference": "foo",
        'image_unique_id': 'deadbeef'}
    agent.testcommand = 'test -f'
    agent.setupcommand = 'test -f'
    agent.provommand = 'test -f'
    monkeypatch.setattr(datetime, 'datetime', FakeDate)
    payload = {'foo': 'bar'}
    os.makedirs(
        os.path.join(
            agent.execution_basedir,
            '42_1970-01-01T00:00:00'),
        exist_ok=True)
    with open(os.path.join(
        agent.execution_basedir,
            '42_1970-01-01T00:00:00',
            'spi_test_result.json'), 'w+') as outf:
            json.dump(payload, outf)
    r = agent.process_test(test_opp)
    assert r == {
        'agent_id': 'agent-queue',
        'image_reference': 'foo',
        'result_payload': payload,
        'test_opportunity_id': '42',
        'test_status': 'PASSED',
        'image_unique_id': 'deadbeef',
        "primary_snap_name": "foo",
        "product": {"primary_snap_name": "foo"},
        'agent': {
            'id': 'agent-queue',
            'timing': {
                'finished': '1970-01-01T00:00:00',
                'provisioning_start': '1970-01-01T00:00:00',
                'setup_start': '1970-01-01T00:00:00',
                'testing_start': '1970-01-01T00:00:00'
            }
        }
    }


def test_process_provfail(agent, monkeypatch):
    """If provisioning fails, status is PROVFAIL."""
    test_opp = {
        "id": "43",
        "product": {"primary_snap_name": "foo"},
        "image_reference": "foo",
        "image_unique_id": "deadbeef"}
    agent.provcommand = '/bin/false'
    monkeypatch.setattr(datetime, 'datetime', FakeDate)
    r = agent.process_test(test_opp)
    assert r == {
        'agent_id': 'agent-queue',
        'image_reference': 'foo',
        'result_payload': {},
        'test_opportunity_id': '43',
        'test_status': 'PROVFAIL',
        'image_unique_id': 'deadbeef',
        'primary_snap_name': 'foo',
        "product": {"primary_snap_name": "foo"},
        'agent': {
            'id': 'agent-queue',
            'timing': {
                'finished': '1970-01-01T00:00:00',
                'provisioning_start': '1970-01-01T00:00:00',
            }
        }
    }


def test_setup_setupfail(agent, monkeypatch):
    """If setup fails, status is SETUPFAIL."""
    test_opp = {
        "id": "4a",
        "product": {"primary_snap_name": "foo"},
        "image_reference": "foo",
        "image_unique_id": "deadbeef"}
    agent.setupcommand = '/bin/false'
    monkeypatch.setattr(datetime, 'datetime', FakeDate)
    r = agent.process_test(test_opp)
    assert r == {
        'agent_id': 'agent-queue',
        'image_reference': 'foo',
        'result_payload': {},
        'test_opportunity_id': '4a',
        'test_status': 'SETUPFAIL',
        'image_unique_id': 'deadbeef',
        "primary_snap_name": "foo",
        "product": {"primary_snap_name": "foo"},
        'agent': {
            'id': 'agent-queue',
            'timing': {
                'finished': '1970-01-01T00:00:00',
                'provisioning_start': '1970-01-01T00:00:00',
                'setup_start': '1970-01-01T00:00:00',
            }
        }
    }


def test_process_testfail(agent, monkeypatch):
    """If test fails, status is FAILED."""
    test_opp = {
        "id": "44",
        "product": {"primary_snap_name": "foo"},
        "image_reference": "foo",
        "image_unique_id": "deadbeef"}
    agent.setupcommand = '/bin/true'
    agent.testcommand = '/bin/false'
    monkeypatch.setattr(datetime, 'datetime', FakeDate)
    r = agent.process_test(test_opp)
    payload = {}
    assert r == {
        'agent_id': 'agent-queue',
        'image_reference': 'foo',
        'result_payload': payload,
        'test_opportunity_id': '44',
        'test_status': 'FAILED',
        'image_unique_id': 'deadbeef',
        'primary_snap_name': 'foo',
        'product': {
            'primary_snap_name': 'foo'
        },
        'agent': {
            'id': 'agent-queue',
            'timing': {
                'finished': '1970-01-01T00:00:00',
                'provisioning_start': '1970-01-01T00:00:00',
                'setup_start': '1970-01-01T00:00:00',
                'testing_start': '1970-01-01T00:00:00'
            }
        }
    }


def test_default_config_print(capsys, agent):
    """Test that it prints default config."""
    agent.confpath = None
    agent.do_default_config()
    data = capsys.readouterr()[0]
    assert '[config]' in data
    assert '[provkit]' in data
    assert '[environment]' in data


def test_verbose(agent):
    """Test that -v sets logging to DEBUG."""
    assert agent.log.getEffectiveLevel() == logging.DEBUG


def test_not_verbose(tmpdir):
    """Test that without -v sets logging level ok."""
    import agent
    _opts = opts.copy()
    _opts['-v'] = False
    _opts['<configfile>'] = str(tmpdir.join('conf'))
    conf = ConfigParser()
    with open(_opts['<configfile>'], 'w') as cf:
        conf.write(cf)
        cf.close()
    a = agent.Agent(_opts)
    assert a.log.getEffectiveLevel() == logging.INFO


def test_start_unregistered_success(agent, monkeypatch):
    """Not registered, should register."""
    register_mock = MagicMock(return_value=True)
    poll_mock = MagicMock(return_value=True)
    process_mock = MagicMock(return_value=True)
    post_mock = MagicMock(return_value=True)
    monkeypatch.setattr(agent, 'register', register_mock)
    monkeypatch.setattr(agent, 'poll', poll_mock)
    monkeypatch.setattr(agent, 'process_test', process_mock)
    monkeypatch.setattr(agent, 'post_result', post_mock)
    r = agent.start()
    assert r is True
    register_mock.assert_called_once_with()
    poll_mock.assert_called_once_with()
    process_mock.assert_called_once_with(True)
    post_mock.assert_called_once_with(True)


def test_start_unregistered_failure(agent, monkeypatch):
    """Not registered, should register, fail."""
    register_mock = MagicMock(return_value=False)
    poll_mock = MagicMock(return_value=True)
    monkeypatch.setattr(agent, 'register', register_mock)
    monkeypatch.setattr(agent, 'poll', poll_mock)
    r = agent.start()
    assert r is False
    register_mock.assert_called_once_with()
    poll_mock.assert_not_called()


def test_start_registered_notest(agent, monkeypatch):
    """Registered, should poll, fail."""
    agent.queue_url = 'http://example.com'
    poll_mock = MagicMock(return_value=False)
    process_mock = MagicMock(return_value=True)
    monkeypatch.setattr(agent, 'poll', poll_mock)
    monkeypatch.setattr(agent, 'process_test', process_mock)
    r = agent.start()
    assert r is False
    poll_mock.assert_called_once_with()
    process_mock.assert_not_called()


def test_start_result_retries(agent, monkeypatch):
    """Test that post_result is retried."""
    register_mock = MagicMock(return_value=True)
    poll_mock = MagicMock(return_value=True)
    process_mock = MagicMock(return_value=True)
    monkeypatch.setattr(agent, 'register', register_mock)
    monkeypatch.setattr(agent, 'poll', poll_mock)
    monkeypatch.setattr(agent, 'process_test', process_mock)
    monkeypatch.setattr(time, 'sleep', MagicMock())

    post_mock = MagicMock()
    post_mock.side_effect = [False, True]
    monkeypatch.setattr(agent, 'post_result', post_mock)
    r = agent.start()
    assert r is True
    assert post_mock.call_count == 2

    post_mock = MagicMock()
    post_mock.side_effect = [False, False, False, False, False, False, True]
    monkeypatch.setattr(agent, 'post_result', post_mock)
    r = agent.start()
    assert r is False
    assert post_mock.call_count == 3


def test_main_start_fails(agent, monkeypatch):
    """Start fails, so it sleeps."""
    sleepmock = MagicMock()
    start_mock = MagicMock(return_value=False)
    monkeypatch.setattr(agent, 'start', start_mock)
    monkeypatch.setattr(time, 'sleep', sleepmock)
    agent.main()
    start_mock.assert_called_once_with()
    sleepmock.assert_called_once_with(agent.interval)


def test_main_start_succeeds(agent, monkeypatch):
    """Start fails, so it sleeps."""
    sleepmock = MagicMock()
    start_mock = MagicMock(return_value=True)
    monkeypatch.setattr(agent, 'start', start_mock)
    monkeypatch.setattr(time, 'sleep', sleepmock)
    agent.main()
    start_mock.assert_called_once_with()
    assert sleepmock.called is False


def test_main_start_interrupted(agent, monkeypatch):
    """KBD interrupt, exit."""
    mock = MagicMock(side_effect=KeyboardInterrupt)
    monkeypatch.setattr(agent, 'start', mock)
    with pytest.raises(SystemExit):
        agent.main()


def test_main_start_exception(agent, monkeypatch):
    """Random exception, wait."""
    sleepmock = MagicMock()
    mock = MagicMock(side_effect=Exception)
    monkeypatch.setattr(agent, 'start', mock)
    monkeypatch.setattr(time, 'sleep', sleepmock)
    agent.main()
    sleepmock.assert_called_once_with(agent.interval)


def test_main_drain(agent, monkeypatch):
    """Drain doesn't wait, exits."""
    agent.drain = True
    sleepmock = False  # Should not be called
    mock = MagicMock(return_value=False)
    monkeypatch.setattr(agent, 'start', mock)
    monkeypatch.setattr(time, 'sleep', sleepmock)
    with pytest.raises(SystemExit):
        agent.main()


def test_makedirs_collision():
    """Fail if we collide with a file."""
    import agent
    with pytest.raises(OSError):
        agent.makedirs('/etc/passwd')


def test_makedirs_exists():
    """Succeed if we collide with a directory."""
    import agent
    agent.makedirs('/etc')


def test_makedirs_succeeds(tmpdir):
    """Succeed if target doesn't exist."""
    import agent
    target = str(tmpdir.join('foo', 'bar'))
    agent.makedirs(target)
    assert os.path.isdir(target)


def test_makedirs_errors(tmpdir, monkeypatch):
    """Re-raise errors."""
    import agent
    mock = MagicMock(side_effect=Exception)
    monkeypatch.setattr(os, 'makedirs', mock)
    target = str(tmpdir.join('foo', 'bar'))
    with pytest.raises(Exception):
        agent.makedirs(target)


def test_dispatch_working_dir(agent, caplog, tmpdir):
    """Test we run in the right folder."""
    log = str(tmpdir.mkdir('sub').join('log'))
    with caplog.atLevel(logging.INFO):
        r = agent.dispatch(
            '',
            '/bin/pwd',
            str(log),
            working_dir=str(tmpdir))
        assert r is True
    assert str(tmpdir) in open(log).read().strip()


def test_dispatch_envvar(agent, caplog, tmpdir, monkeypatch):
    """Test we run in the right folder."""
    log = str(tmpdir.mkdir('sub').join('log'))
    monkeypatch.setenv('FOO', 'BAR')
    with caplog.atLevel(logging.INFO):
        r = agent.dispatch(
            '',
            'echo $FOO',
            str(log),
            working_dir=str(tmpdir))
        assert r is True
    assert 'BAR' in open(log).read().strip()


def test_config_envvar(agent, caplog, tmpdir, monkeypatch):
    """Test we pass the environment section in the scripts env."""
    log = str(tmpdir.mkdir('sub').join('log'))
    agent.config.add_section('environment')
    agent.config.set('environment', 'FOO', 'BAR')
    with caplog.atLevel(logging.INFO):
        r = agent.dispatch(
            '',
            'echo $FOO',
            str(log),
            working_dir=str(tmpdir))
        assert r is True
    assert 'BAR' in open(log).read().strip()


def test_process_provision_workdir(agent, monkeypatch, caplog, tmpdir):
    """Ensure we are always passing working_dir to dispatch."""
    test_opp = {
        "id": "40b",
        "product": {"primary_snap_name": "foo"},
        "image_reference": "foo",
        "image_unique_id": "deadbeef"}
    agent.execution_basedir = str(tmpdir)
    agent.testcommand = '/bin/echo testing now'
    dispatchmock = MagicMock()
    rmtreemock = MagicMock()
    monkeypatch.setattr(agent, 'dispatch', dispatchmock)
    monkeypatch.setattr(datetime, 'datetime', FakeDate)
    monkeypatch.setattr(shutil, 'rmtree', rmtreemock)
    agent.process_test(test_opp)
    dispatchmock.assert_has_calls([
        call(
            ANY,
            '/bin/echo provisioning now',
            ANY,
            os.path.join(
                agent.execution_basedir,
                '40b_1970-01-01T00:00:00')),
        call(
            ANY,
            '/bin/echo setting up now',
            ANY,
            os.path.join(
                agent.execution_basedir,
                '40b_1970-01-01T00:00:00')),
        call(
            ANY,
            '/bin/echo testing now',
            ANY,
            os.path.join(
                agent.execution_basedir,
                '40b_1970-01-01T00:00:00'))
    ], True)
    rmtreemock.assert_called_once_with(
        os.path.join(
            agent.execution_basedir,
            '40b_1970-01-01T00:00:00'))


def test_process_provision_workdir_debug(agent, monkeypatch, caplog, tmpdir):
    """Ensure we are always passing working_dir to dispatch."""
    test_opp = {
        "id": "40b",
        "image_reference": "foo",
        "product": {"primary_snap_name": "foo"},
        "image_unique_id": "deadbeef"}
    agent.execution_basedir = str(tmpdir)
    agent.debug_mode = True
    agent.testcommand = '/bin/echo testing now'
    dispatchmock = MagicMock()
    monkeypatch.setattr(agent, 'dispatch', dispatchmock)
    monkeypatch.setattr(datetime, 'datetime', FakeDate)
    monkeypatch.setattr(shutil, 'rmtree', 'notcallable')
    agent.process_test(test_opp)
    dispatchmock.assert_has_calls([
        call(
            ANY,
            '/bin/echo provisioning now',
            ANY,
            os.path.join(
                agent.execution_basedir,
                'debug')),
        call(
            ANY,
            '/bin/echo setting up now',
            ANY,
            os.path.join(
                agent.execution_basedir,
                'debug')),
        call(
            ANY,
            '/bin/echo testing now',
            ANY,
            os.path.join(
                agent.execution_basedir,
                'debug'))
    ], True)


def test_interactive_login(agent, monkeypatch, capsys):
    """Test interactive login feature."""
    agent.confpath = None
    result = MagicMock()
    result.status_code = 200
    result.json.return_value = {
        'consumer_key': 'consumer_key',
        'consumer_secret': 'consumer_secret',
        'token_key': 'token_key',
        'token_secret': 'token_secret',
    }
    post_mock = MagicMock()
    post_mock.return_value = result
    monkeypatch.setattr(requests, 'post', post_mock)
    monkeypatch.setattr(builtins, 'input', lambda *a: 'foo')
    monkeypatch.setattr(getpass, 'getpass', lambda *a: 'bar')
    agent.do_interactive_login()
    assert '[auth]' in capsys.readouterr()[0]
    post_mock.assert_called_once_with(
        'https://login.ubuntu.com/api/v2/tokens/oauth',
        timeout=10,
        json={'token_name': 'agent-queue', 'email': 'foo', 'password': 'bar'}
    )


def test_sso_envvar(agent, monkeypatch, capsys):
    """Test interactive login feature with custom SSO URL."""
    agent.confpath = None
    monkeypatch.setenv('SSO_BASE_URL', 'http://foo.com')
    result = MagicMock()
    result.status_code = 200
    result.json.return_value = {
        'consumer_key': 'consumer_key',
        'consumer_secret': 'consumer_secret',
        'token_key': 'token_key',
        'token_secret': 'token_secret',
    }
    post_mock = MagicMock()
    post_mock.return_value = result
    monkeypatch.setattr(requests, 'post', post_mock)
    monkeypatch.setattr(builtins, 'input', lambda *a: 'foo')
    monkeypatch.setattr(getpass, 'getpass', lambda *a: 'bar')
    agent.do_interactive_login()
    assert '[auth]' in capsys.readouterr()[0]
    post_mock.assert_called_once_with(
        'http://foo.com/api/v2/tokens/oauth',
        timeout=10,
        json={'token_name': 'agent-queue', 'password': 'bar', 'email': 'foo'}
    )


def test_interactive_login_2fa(agent, monkeypatch, capsys):
    """Test interactive login feature."""
    agent.confpath = None

    req_2fa = MagicMock()
    req_2fa.status_code = 401
    req_2fa.json.return_value = {
        'code': 'TWOFACTOR_REQUIRED'
    }

    result = MagicMock()
    result.status_code = 200
    result.json.return_value = {
        'consumer_key': 'consumer_key',
        'consumer_secret': 'consumer_secret',
        'token_key': 'token_key',
        'token_secret': 'token_secret',
    }

    postmock = MagicMock()
    postmock.side_effect = [req_2fa, result]

    monkeypatch.setattr(requests, 'post', postmock)
    monkeypatch.setattr(builtins, 'input', lambda *a: 'foo')
    monkeypatch.setattr(getpass, 'getpass', lambda *a: 'bar')
    agent.do_interactive_login()
    assert '[auth]' in capsys.readouterr()[0]


def test_oauth_session(monkeypatch, tmpdir):
    """Test that Oauth session is used if auth is configured."""
    cf = ConfigParser()
    cf['auth'] = {
        'consumer_key': 'foo',
        'consumer_secret': 'bar',
        'token_key': 'baz',
        'token_secret': 'bat',
    }
    import agent
    _opts = opts.copy()
    _opts['-v'] = False
    _opts['<configfile>'] = str(tmpdir.join('conf'))
    with open(_opts['<configfile>'], 'w') as fd:
        cf.write(fd)
    session_mock = MagicMock()
    monkeypatch.setattr(requests_oauthlib, 'OAuth1Session', session_mock)
    agent = agent.Agent(_opts)
    session_mock.assert_called_once_with(
        'foo',
        client_secret='bar',
        resource_owner_key='baz',
        resource_owner_secret='bat',
    )


def test_no_oauth_session(monkeypatch, tmpdir):
    """Test that Oauth session is not used if auth is not configured."""
    import agent
    _opts = opts.copy()
    _opts['-v'] = False
    _opts['<configfile>'] = str(tmpdir.join('conf'))
    conf = ConfigParser()
    conf['config'] = agent.DEFAULT_CONFIG
    with open(_opts['<configfile>'], 'w') as fd:
        conf.write(fd)
        fd.close()
    session_mock = MagicMock()
    monkeypatch.setattr(requests_oauthlib, 'OAuth1Session', session_mock)
    agent = agent.Agent(_opts)
    assert not session_mock.called


def test_empty_value_warning(request, caplog, monkeypatch):
    """Test that warn_empty is called correctly."""
    import agent as agent_mod
    mock = MagicMock()
    monkeypatch.setattr(agent_mod.Agent, 'warn_empty_config_values', mock)
    agent(request)
    mock.assert_called_once_with()
